import pathlib
from typing import Dict, Union

import geopandas as gpd
import numpy as np
import rioxarray  # noqa # pylint: disable=unused-import
import xarray as xr

import gistim


def write_raster(
    head: xr.DataArray,
    crs: int,
    outpath: Union[pathlib.Path, str],
) -> None:
    out = head.rio.write_crs(crs).rio.write_coordinate_system()
    out.to_netcdf(outpath.with_suffix(".nc"))
    return


def write_ugrid(
    head: xr.DataArray,
    crs: int,
    outpath: Union[pathlib.Path, str],
) -> None:
    ugrid_head = gistim.to_ugrid2d(head)
    ugrid_head["projected_coordinate_system"] = xr.DataArray(
        data=np.int32(0),
        attrs={"epsg": np.int32(crs.to_epsg())},
    )
    ugrid_head.to_netcdf(outpath.with_suffix(".ugrid.nc"))
    return


def write_vector(
    gdf_head: gpd.GeoDataFrame,
    crs: int,
    outpath: Union[pathlib.Path, str],
    layername: str,
) -> None:
    if len(gdf_head.index) > 0:
        gdf_head = gdf_head.set_crs(crs)
        gdf_head.to_file(
            outpath.with_suffix(".output.gpkg"),
            driver="GPKG",
            layer=layername,
        )
    return


def compute_steady(
    inpath: Union[pathlib.Path, str],
    outpath: Union[pathlib.Path, str],
    cellsize: float,
    active_elements: Dict[str, bool],
) -> None:
    inpath = pathlib.Path(inpath)
    outpath = pathlib.Path(outpath)

    timml_spec, _ = gistim.model_specification(inpath, active_elements)
    timml_model, _, observations = gistim.timml_elements.initialize_model(timml_spec)
    timml_model.solve()

    extent, crs = gistim.gridspec(inpath, cellsize)
    gdfs_obs = gistim.timml_elements.head_observations(timml_model, observations)
    head = gistim.timml_elements.headgrid(timml_model, extent, cellsize)

    write_raster(head, crs, outpath)
    write_ugrid(head, crs, outpath)

    for name, gdf in gdfs_obs.items():
        write_vector(gdf, crs, outpath, layername=name)

    return


def compute_transient(
    inpath: Union[pathlib.Path, str],
    outpath: Union[pathlib.Path, str],
    cellsize: float,
    active_elements: Dict[str, bool],
) -> None:
    inpath = pathlib.Path(inpath)
    outpath = pathlib.Path(outpath)

    timml_spec, ttim_spec = gistim.model_specification(inpath, active_elements)
    timml_model, _, observations = gistim.timml_elements.initialize_model(timml_spec)
    ttim_model, _, observations = gistim.ttim_elements.initialize_model(
        ttim_spec, timml_model
    )
    timml_model.solve()
    ttim_model.solve()

    extent, crs = gistim.gridspec(inpath, cellsize)
    refdate = ttim_spec.temporal_settings["reference_date"].iloc[0]
    gdfs_obs = gistim.ttim_elements.head_observations(ttim_model, refdate, observations)

    # If no output times are specified, just compute the TimML steady-state
    # heads.
    if len(ttim_spec.output_times) > 0:
        head = gistim.ttim_elements.headgrid(
            ttim_model,
            extent,
            cellsize,
            ttim_spec.output_times,
            refdate,
        )
        write_raster(head, crs, outpath)
        write_ugrid(head, crs, outpath)
    else:
        head = gistim.timml_elements.headgrid(timml_model, extent, cellsize)

    write_raster(head, crs, outpath)
    write_ugrid(head, crs, outpath)

    for name, gdf in gdfs_obs.items():
        write_vector(gdf, crs, outpath, layername=name)

    return



"""
This functions in this module extract data from a (specially prepared) netCDF
for a given polygon and compute a number of statistics, which is presented in
table form.

These statistics can also be formatted so they maybe pasted into QGIS, as a
ready made aquifer-aquitard parametrization.
"""
import pathlib
from typing import List, Union

import affine
import geopandas as gpd
import numpy as np
import pandas as pd
import rasterio.features
import shapely.wkt
import xarray as xr


def coord_reference(da_coord):
    """
    Extracts dx, xmin, xmax for a coordinate DataArray, where x is any coordinate.

    If the DataArray coordinates are nonequidistant, dx will be returned as
    1D ndarray instead of float.

    Parameters
    ----------
    a : xarray.DataArray of a coordinate

    Returns
    --------------
    tuple
        (dx, xmin, xmax) for a coordinate x
    """
    x = da_coord.values

    # Possibly non-equidistant
    dx_string = f"d{da_coord.name}"
    if dx_string in da_coord.coords:
        dx = da_coord.coords[dx_string]
        if (dx.shape == x.shape) and (dx.size != 1):
            # choose correctly for decreasing coordinate
            if dx[0] < 0.0:
                end = 0
                start = -1
            else:
                start = 0
                end = -1
            dx = dx.values.astype(np.float64)
            xmin = float(x.min()) - 0.5 * abs(dx[start])
            xmax = float(x.max()) + 0.5 * abs(dx[end])
            # As a single value if equidistant
            if np.allclose(dx, dx[0]):
                dx = dx[0]
        else:
            dx = float(dx)
            xmin = float(x.min()) - 0.5 * abs(dx)
            xmax = float(x.max()) + 0.5 * abs(dx)
    elif x.size == 1:
        raise ValueError(
            f"DataArray has size 1 along {da_coord.name}, so cellsize must be provided"
            f" as a coordinate named d{da_coord.name}."
        )
    else:  # Equidistant
        # TODO: decide on decent criterium for what equidistant means
        # make use of floating point epsilon? E.g:
        # https://github.com/ioam/holoviews/issues/1869#issuecomment-353115449
        dxs = np.diff(x.astype(np.float64))
        dx = dxs[0]
        atolx = abs(1.0e-4 * dx)
        if not np.allclose(dxs, dx, atolx):
            raise ValueError(
                f"DataArray has to be equidistant along {da_coord.name}, or cellsizes"
                f" must be provided as a coordinate named d{da_coord.name}."
            )

        # as xarray uses midpoint coordinates
        xmin = float(x.min()) - 0.5 * abs(dx)
        xmax = float(x.max()) + 0.5 * abs(dx)

    return dx, xmin, xmax


def spatial_reference(a):
    """
    Extracts spatial reference from DataArray.

    If the DataArray coordinates are nonequidistant, dx and dy will be returned
    as 1D ndarray instead of float.

    Parameters
    ----------
    a : xarray.DataArray

    Returns
    --------------
    tuple
        (dx, xmin, xmax, dy, ymin, ymax)

    """
    dx, xmin, xmax = coord_reference(a["x"])
    dy, ymin, ymax = coord_reference(a["y"])
    return dx, xmin, xmax, dy, ymin, ymax


def transform(a):
    """
    Extract the spatial reference information from the DataArray coordinates,
    into an affine.Affine object for writing to e.g. rasterio supported formats.

    Parameters
    ----------
    a : xarray.DataArray

    Returns
    -------
    affine.Affine

    """
    dx, xmin, _, dy, _, ymax = spatial_reference(a)

    def equidistant(dx, name):
        if isinstance(dx, np.ndarray):
            if np.unique(dx).size == 1:
                return dx[0]
            else:
                raise ValueError(f"DataArray is not equidistant along {name}")
        else:
            return dx

    dx = equidistant(dx, "x")
    dy = equidistant(dy, "y")

    if dx < 0.0:
        raise ValueError("dx must be positive")
    if dy > 0.0:
        raise ValueError("dy must be negative")
    return affine.Affine(dx, 0.0, xmin, 0.0, dy, ymax)


def rasterize(geodataframe, like, fill, **kwargs):
    """
    Rasterize a geopandas GeoDataFrame onto the given
    xarray coordinates.

    Parameters
    ----------
    geodataframe : geopandas.GeoDataFrame
    column : str, int, float
        column name of geodataframe to burn into raster
    like : xarray.DataArray
        Example DataArray. The rasterized result will match the shape and
        coordinates of this DataArray.
    fill : float, int
        Fill value for nodata areas. Optional, default value is np.nan.
    kwargs : additional keyword arguments for rasterio.features.rasterize.
        See: https://rasterio.readthedocs.io/en/stable/api/rasterio.features.html#rasterio.features.rasterize

    Returns
    -------
    rasterized : xarray.DataArray
        Vector data rasterized. Matches shape and coordinates of ``like``.
    """
    shapes = [geom for geom in geodataframe.geometry]
    # shapes must be an iterable
    try:
        iter(shapes)
    except TypeError:
        shapes = (shapes,)

    raster = rasterio.features.rasterize(
        shapes,
        out_shape=like.shape,
        fill=fill,
        transform=transform(like),
        **kwargs,
    )
    return xr.DataArray(raster, like.coords, like.dims)


def layer_statistics(path: Union[pathlib.Path, str], gdf: gpd.GeoDataFrame):
    """
    Extract summary statistics of all variables present in a Dataset.

    Variables must contain dimensions ``("layer", "y","x")``.

    Parameters
    ----------
    path: Union[pathlib.Path, str]
        Path to the netCDF dataset containing geohydrologic subsoil properties.
    geodataframe: gpd.GeoDataFrame
        Defines area of interest

    Returns
    -------
    statistics: pandas.DataFrame
        The layers are stored as rows, the variables as columns.
    """
    xmin, ymin, xmax, ymax = gdf.total_bounds
    ds = xr.open_dataset(path).sel(x=slice(xmin, xmax), y=slice(ymax, ymin))
    first_var = next(iter(ds.data_vars.keys()))
    like = ds[first_var].isel(layer=0).drop("layer")
    area_of_interest = rasterize(gdf, like, fill=0) != 0

    stats = pd.DataFrame()
    for variable in ds.data_vars:
        da = ds[variable].where(area_of_interest)
        stats[f"{variable}-min"] = da.min(["y", "x"])  # .to_dataframe(),
        stats[f"{variable}-max"] = da.max(["y", "x"])  # .to_dataframe(),
        stats[f"{variable}-mean"] = da.mean(["y", "x"])  # .to_dataframe(),
        stats[f"{variable}-median"] = da.median(["y", "x"])  # .to_dataframe(),
        stats[f"{variable}-std"] = da.std(["y", "x"])  # .to_dataframe(),
    return stats


def as_aquifer_aquitard(dataframe: pd.DataFrame, statistic: str = "mean"):
    """
    Convert a layer statistics dataframe into a table that can be directly
    copy-pasted into the QGIS plugin Aquifer properties layer.

    Primarily, this means interleaving conductivities and resistances
    in separate rows.

    Parameters
    ----------
    dataframe: pandas.DataFrame
        A dataframe containing layer statistics.
    statistic: str, optional
        Which statistic to use. Default value: ``"mean"``

    Returns
    -------
    aquifer_aquitard_table: pandas.DataFrame
        Reshuffled rows, containing only the data required for the plugin
        Aquifer properties.
    """
    n_layer = len(dataframe)
    out = pd.DataFrame()
    out["fid"] = np.arange(n_layer)
    out["layer"] = np.arange(n_layer)
    out["aquitard_c"] = np.nan
    out["aquitard_c"].values[1:] = dataframe[f"c-{statistic}"].values[:-1]
    out["aquitard_npor"] = np.nan
    out["aquitard_s"] = np.nan
    out["aquifer_k"] = dataframe[f"kh-{statistic}"].values
    out["aquifer_npor"] = np.nan
    out["aquifer_s"] = np.nan
    out["aquifer_top"] = dataframe[f"top-{statistic}"]
    out["aquifer_bottom"] = dataframe[f"bottom-{statistic}"]
    out["semiconf_top"] = np.nan
    out["semiconf_head"] = np.nan
    return out


def netcdf_to_table(inpath: str, outpath: str, wkt_geometry: List[str]) -> None:
    geometries = [shapely.wkt.loads(geom) for geom in wkt_geometry]
    outpath = pathlib.Path(outpath)
    gdf = gpd.GeoDataFrame(geometry=geometries)
    stats = layer_statistics(inpath, gdf)
    stats.to_csv(outpath.parent / f"{outpath.stem}-statistics.csv")
    table = as_aquifer_aquitard(stats)
    table.to_csv(outpath)



"""
This model converts the geodataframe as read from the geopackage into keyword
arguments for TimML.

These keyword arguments are used to initialize a model, or used to generate a
Python script.
"""
from collections import defaultdict
from typing import Any, Dict, List, Tuple

import black
import geopandas as gpd
import numpy as np
import timml
import tqdm
import xarray as xr

from gistim.common import (
    ElementSpecification,
    TimmlModelSpecification,
    aquifer_data,
    dict_to_kwargs_code,
    filter_nan,
    headgrid_code,
    linestring_coordinates,
    point_coordinates,
    polygon_coordinates,
    sanitized,
)


def constant(spec: ElementSpecification) -> List[Dict[str, Any]]:
    firstrow = filter_nan(spec.dataframe.iloc[0])
    x, y = point_coordinates(firstrow)
    return [
        {
            "xr": x,
            "yr": y,
            "hr": firstrow["head"],
            "layer": firstrow["layer"],
            "label": firstrow["label"],
        }
    ]


def uflow(spec: ElementSpecification) -> List[Dict[str, Any]]:
    row = filter_nan(spec.dataframe.iloc[0])
    return [
        {
            "slope": row["slope"],
            "angle": row["angle"],
            "label": row["label"],
        }
    ]


def observation(spec: ElementSpecification) -> List[Dict[str, Any]]:
    dataframe = spec.dataframe
    X, Y = point_coordinates(dataframe)
    kwargslist = []
    kwargslist = []
    for (row, x, y) in zip(dataframe.to_dict("records"), X, Y):
        row = filter_nan(row)
        kwargslist.append(
            {
                "x": x,
                "y": y,
                "label": row["label"],
            }
        )
    return kwargslist


def well(spec: ElementSpecification) -> List[Dict[str, Any]]:
    dataframe = spec.dataframe
    X, Y = point_coordinates(dataframe)
    kwargslist = []
    for (row, x, y) in zip(dataframe.to_dict("records"), X, Y):
        row = filter_nan(row)
        kwargslist.append(
            {
                "xw": x,
                "yw": y,
                "Qw": row["discharge"],
                "rw": row["radius"],
                "res": row["resistance"],
                "layers": row["layer"],
                "label": row["label"],
            }
        )
    return kwargslist


def headwell(spec: ElementSpecification) -> List[Dict[str, Any]]:
    dataframe = spec.dataframe
    X, Y = point_coordinates(dataframe)
    kwargslist = []
    for (row, x, y) in zip(dataframe.to_dict("records"), X, Y):
        row = filter_nan(row)
        kwargslist.append(
            {
                "xw": x,
                "yw": y,
                "hw": row["head"],
                "rw": row["radius"],
                "res": row["resistance"],
                "layers": row["layer"],
                "label": row["label"],
            }
        )
    return kwargslist


def polygoninhom(spec: ElementSpecification) -> List[Dict[str, Any]]:
    """
    Parameters
    ----------
    dataframe: tuple of geopandas.GeoDataFrame

    Returns
    -------
    None
    """
    geometry = spec.dataframe
    properties = spec.associated_dataframe.set_index("inhomogeneity_id")
    # Iterate through the row containing the geometry
    # and iterate through the associated table containing k properties.
    kwarglist = []
    for row in geometry.to_dict("records"):
        row = filter_nan(row)
        dataframe = properties.loc[[row["inhomogeneity_id"]]]
        kwargs = aquifer_data(dataframe)
        kwargs["xy"] = polygon_coordinates(row)
        kwargs["order"] = row["order"]
        kwargs["ndeg"] = row["ndegrees"]
        kwarglist.append(kwargs)
    return kwarglist


def polygonareasink(spec: ElementSpecification) -> List[Dict[str, Any]]:
    geometry = spec.dataframe
    properties = spec.associated_dataframe.sort_values(by="layer").set_index("layer")
    # Ensure there's no semiconfined top
    dataframe = properties.copy()
    dataframe.loc[0, "semiconf_head"] = np.nan
    kwarglist = []
    for row in geometry.to_dict("records"):
        row = filter_nan(row)
        kwargs = aquifer_data(dataframe.reset_index())
        kwargs["N"] = row["rate"]
        kwargs["xy"] = polygon_coordinates(row)
        kwargs["order"] = row["order"]
        kwargs["ndeg"] = row["ndegrees"]
        kwarglist.append(kwargs)
    return kwarglist


def polygontop(spec: ElementSpecification) -> List[Dict[str, Any]]:
    geometry = spec.dataframe
    properties = spec.associated_dataframe.sort_values(by="layer").set_index("layer")
    kwarglist = []
    for row in geometry.to_dict("records"):
        row = filter_nan(row)
        dataframe = properties.copy()
        dataframe.loc[0, "aquitard_c"] = row["aquitard_c"]
        dataframe.loc[0, "semiconf_top"] = row["semiconf_top"]
        dataframe.loc[0, "semiconf_head"] = row["semiconf_head"]
        kwargs = aquifer_data(dataframe.reset_index())
        kwargs["xy"] = polygon_coordinates(row)
        kwargs["order"] = row["order"]
        kwargs["ndeg"] = row["ndegrees"]
        kwarglist.append(kwargs)
    return kwarglist


def buildingpit(spec: ElementSpecification) -> List[Dict[str, Any]]:
    geometry = spec.dataframe
    properties = spec.associated_dataframe.set_index("timeseries_id")
    # Iterate through the row containing the geometry
    # and iterate through the associated table containing k properties.
    kwarglist = []
    for row in geometry.to_dict("records"):
        row = filter_nan(row)
        dataframe = properties.loc[row["timeseries_id"]]
        kwargs = aquifer_data(dataframe)
        kwargs["xy"] = polygon_coordinates(row)
        kwargs["order"] = row["order"]
        kwargs["ndeg"] = row["ndegrees"]
        kwargs["layers"] = np.atleast_1d(row["layer"])
        kwarglist.append(kwargs)
    return kwarglist


def headlinesink(spec: ElementSpecification) -> List[Dict[str, Any]]:
    kwargslist = []
    for row in spec.dataframe.to_dict("records"):
        row = filter_nan(row)
        kwargslist.append(
            {
                "xy": linestring_coordinates(row),
                "hls": row["head"],
                "res": row["resistance"],
                "wh": row["width"],
                "order": row["order"],
                "layers": row["layer"],
                "label": row["label"],
            }
        )
    return kwargslist


def linesinkditch(spec: ElementSpecification) -> List[Dict[str, Any]]:
    kwargslist = []
    for row in spec.dataframe.to_dict("records"):
        row = filter_nan(row)
        kwargslist.append(
            {
                "xy": linestring_coordinates(row),
                "Qls": row["discharge"],
                "res": row["resistance"],
                "wh": row["width"],
                "order": row["order"],
                "layers": row["layer"],
                "label": row["label"],
            }
        )
    return kwargslist


def leakylinedoublet(spec: ElementSpecification) -> List[Dict[str, Any]]:
    kwargslist = []
    for row in spec.dataframe.to_dict("records"):
        row = filter_nan(row)
        kwargslist.append(
            {
                "xy": linestring_coordinates(row),
                "res": row["resistance"],
                "layers": row["layer"],
                "order": row["order"],
                "label": row["label"],
            }
        )
    return kwargslist


def implinedoublet(spec: ElementSpecification) -> List[Dict[str, Any]]:
    kwargslist = []
    for row in spec.dataframe.to_dict("records"):
        row = filter_nan(row)
        kwargslist.append(
            {
                "xy": linestring_coordinates(row),
                "layers": row["layer"],
                "order": row["order"],
                "label": row["label"],
            }
        )
    return kwargslist


def circareasink(spec: ElementSpecification) -> List[Dict[str, Any]]:
    kwargslist = []
    for row in spec.dataframe.to_dict("records"):
        row = filter_nan(row)
        xc, yc = np.array(row["geometry"].centroid.coords)[0]
        coords = np.array(row["geometry"].exterior.coords)
        x, y = coords.T
        # Use squared radii
        radii2 = (x - xc) ** 2 + (y - yc) ** 2
        radius2 = radii2[0]
        # Check whether geometry is close enough to a circle.
        # Accept 1% deviation.
        tolerance = 0.01 * radius2
        if not np.allclose(radii2, radius2, atol=tolerance):
            raise ValueError("Circular Area Sink geometry is not circular")

        radius = np.sqrt(radius2)
        kwargslist.append(
            {
                "xc": xc,
                "yc": yc,
                "R": radius,
                "N": row["rate"],
            }
        )
    return kwargslist


def validate(spec: TimmlModelSpecification) -> None:
    if spec.aquifer is None:
        raise ValueError("Aquifer entry is missing")
    # TODO: more checks


def head_observations(model: timml.Model, observations: Dict) -> gpd.GeoDataFrame:
    geodataframes = {}
    for name, kwargs_collection in observations.items():
        heads = []
        xx = []
        yy = []
        labels = []
        for kwargs in kwargs_collection:
            x = kwargs["x"]
            y = kwargs["y"]
            heads.append(model.head(x=x, y=y))
            xx.append(x)
            yy.append(y)
            labels.append(kwargs["label"])

        d = {"geometry": gpd.points_from_xy(xx, yy), "label": labels}
        for i, layerhead in enumerate(np.vstack(heads).T):
            d[f"head_layer{i}"] = layerhead

        geodataframes[name] = gpd.GeoDataFrame(d)

    return geodataframes


def headgrid(model: timml.Model, extent: Tuple[float], cellsize: float) -> xr.DataArray:
    """
    Compute the headgrid of the TimML model, and store the results
    in an xarray DataArray with the appropriate dimensions.

    Parameters
    ----------
    model: timml.Model
        Solved model to get heads from
    extent: Tuple[float]
        xmin, xmax, ymin, ymax
    cellsize: float
        Desired cell size of the output head grids

    Returns
    -------
    head: xr.DataArray
        DataArray with dimensions ``("layer", "y", "x")``.
    """
    xmin, xmax, ymin, ymax = extent
    x = np.arange(xmin, xmax, cellsize) + 0.5 * cellsize
    # In geospatial rasters, y is DECREASING with row number
    y = np.arange(ymax, ymin, -cellsize) - 0.5 * cellsize
    nlayer = model.aq.find_aquifer_data(x[0], y[0]).naq
    layer = [i for i in range(nlayer)]
    head = np.empty((nlayer, y.size, x.size), dtype=np.float64)
    for i in tqdm.tqdm(range(y.size)):
        for j in range(x.size):
            head[:, i, j] = model.head(x[j], y[i], layer)

    return xr.DataArray(
        data=head,
        name="head",
        coords={"layer": layer, "y": y, "x": x},
        dims=("layer", "y", "x"),
    )


# Map the names of the elements to their constructors
MAPPING = {
    "Constant": (constant, timml.Constant),
    "Uniform Flow": (uflow, timml.Uflow),
    "Circular Area Sink": (circareasink, timml.CircAreaSink),
    "Well": (well, timml.Well),
    "Head Well": (headwell, timml.HeadWell),
    "Polygon Inhomogeneity": (polygoninhom, timml.PolygonInhomMaq),
    "Polygon Area Sink": (polygonareasink, timml.PolygonInhomMaq),
    "Polygon Semi-Confined Top": (polygontop, timml.PolygonInhomMaq),
    "Head Line Sink": (headlinesink, timml.HeadLineSinkString),
    "Line Sink Ditch": (linesinkditch, timml.LineSinkDitchString),
    "Leaky Line Doublet": (leakylinedoublet, timml.LeakyLineDoubletString),
    "Impermeable Line Doublet": (implinedoublet, timml.ImpLineDoubletString),
    "Building Pit": (buildingpit, timml.BuildingPit),
    "Observation": (observation, None),
}


def initialize_model(spec: TimmlModelSpecification) -> timml.Model:
    """
    Initialize a TimML analytic model based on the data in a geopackage.

    Parameters
    ----------
    spec: ModelSpecification
        Named tuple with the layer name of the aquifer and a dictionary of
        other element names to its construction function.

    Returns
    -------
    timml.Model
    elements: Dict of number name to TimML element
    observations: Dict of list

    Examples
    --------

    >>> import gistim
    >>> path = "my-model.gpkg"
    >>> spec = gistim.model_specification(path)
    >>> model = gistim.initialize_model(spec)
    >>> model.solve()

    """
    validate(spec)
    model = timml.ModelMaq(**aquifer_data(spec.aquifer))
    elements = {}
    observations = defaultdict(list)
    for name, element_spec in spec.elements.items():
        if (not element_spec.active) or (len(element_spec.dataframe.index) == 0):
            continue

        elementtype = element_spec.elementtype
        if elementtype not in MAPPING:
            msg = (
                f'Invalid element specification "{elementtype}". '
                f'Available types are: {", ".join(MAPPING.keys())}.'
            )
            raise KeyError(msg)

        f_to_kwargs, element = MAPPING[elementtype]
        for i, kwargs in enumerate(f_to_kwargs(element_spec)):
            if elementtype == "Observation":
                observations[name].append(kwargs)
            else:
                kwargs["model"] = model
                elements[f"{name}_{i}"] = element(**kwargs)

    return model, elements, observations


def convert_to_script(spec: TimmlModelSpecification) -> str:
    """
    Convert model specification to an equivalent Python script.
    """
    modelkwargs = dict_to_kwargs_code(aquifer_data(spec.aquifer))

    observations = {}
    strings = [
        "import numpy as np",
        "import timml",
        f"model = timml.ModelMaq({modelkwargs})",
    ]
    for name, element_spec in spec.elements.items():
        elementtype = element_spec.elementtype

        if elementtype not in MAPPING:
            msg = (
                f'Invalid element specification "{elementtype}". '
                f'Available types are: {", ".join(MAPPING.keys())}.'
            )
            raise KeyError(msg)

        f_to_kwargs, element = MAPPING[elementtype]
        for i, kwargs in enumerate(f_to_kwargs(element_spec)):
            if elementtype == "Observation":
                kwargs.pop("label")
                kwargs = dict_to_kwargs_code(kwargs)
                observations[f"observation_{sanitized(name)}_{i}"] = kwargs
            else:
                kwargs["model"] = "model"
                kwargs = dict_to_kwargs_code(kwargs)
                strings.append(
                    f"{sanitized(name)}_{i} = timml.{element.__name__}({kwargs})"
                )

    strings.append("model.solve()")

    xg, yg = headgrid_code(spec.domain)
    strings.append(f"head = model.headgrid(xg={xg}, yg={yg})")

    # Add all the individual observation points
    for name, kwargs in observations.items():
        strings.append(f"{name} = model.head({kwargs})")

    return black.format_str("\n".join(strings), mode=black.FileMode())



"""
Functions to create UGRID compliant netCDFs that can be read by QGIS via MDAL.
While regular structured netCDFs are also supported by QGIS, especially the
temporal controller works a lot better via MDAL.

This does come at some cost: the UGRID netCDFs are more complicated, and MDAL
can be a little frail. Furthermore, contouring for unstructured grids uses a
different algorithm which generates minor "knickpoints" which do not occur with
the contouring algorithm for structured data (likely because interpolation is
simpler). However, sliding through different timesteps works out of the box.
"""
from typing import Union

import numpy as np
import xarray as xr

FloatArray = np.ndarray
IntArray = np.ndarray


def _check_monotonic(dxs, dim):
    # use xor to check if one or the other
    if not ((dxs > 0.0).all() ^ (dxs < 0.0).all()):
        raise ValueError(f"{dim} is not only increasing or only decreasing")


def _coord(da, dim):
    """
    Transform N xarray midpoints into N + 1 vertex edges
    """
    delta_dim = "d" + dim  # e.g. dx, dy, dz, etc.

    # If empty array, return empty
    if da[dim].size == 0:
        return np.array(())

    if delta_dim in da.coords:  # equidistant or non-equidistant
        dx = da[delta_dim].values
        if dx.shape == () or dx.shape == (1,):  # scalar -> equidistant
            dxs = np.full(da[dim].size, dx)
        else:  # array -> non-equidistant
            dxs = dx
        _check_monotonic(dxs, dim)

    else:  # undefined -> equidistant
        if da[dim].size == 1:
            raise ValueError(
                f"DataArray has size 1 along {dim}, so cellsize must be provided"
                " as a coordinate."
            )
        dxs = np.diff(da[dim].values)
        dx = dxs[0]
        atolx = abs(1.0e-4 * dx)
        if not np.allclose(dxs, dx, atolx):
            raise ValueError(
                f"DataArray has to be equidistant along {dim}, or cellsizes"
                " must be provided as a coordinate."
            )
        dxs = np.full(da[dim].size, dx)

    dxs = np.abs(dxs)
    x = da[dim].values
    if not da.indexes[dim].is_monotonic_increasing:
        x = x[::-1]
        dxs = dxs[::-1]

    # This assumes the coordinate to be monotonic increasing
    x0 = x[0] - 0.5 * dxs[0]
    x = np.full(dxs.size + 1, x0)
    x[1:] += np.cumsum(dxs)
    return x


def _ugrid2d_dataset(
    node_x: FloatArray,
    node_y: FloatArray,
    face_x: FloatArray,
    face_y: FloatArray,
    face_nodes: IntArray,
) -> xr.Dataset:
    ds = xr.Dataset()
    ds["mesh2d"] = xr.DataArray(
        data=0,
        attrs={
            "cf_role": "mesh_topology",
            "long_name": "Topology data of 2D mesh",
            "topology_dimension": 2,
            "node_coordinates": "node_x node_y",
            "face_node_connectivity": "face_nodes",
            "edge_node_connectivity": "edge_nodes",
        },
    )
    ds = ds.assign_coords(
        node_x=xr.DataArray(
            data=node_x,
            dims=["node"],
        )
    )
    ds = ds.assign_coords(
        node_y=xr.DataArray(
            data=node_y,
            dims=["node"],
        )
    )
    ds["face_nodes"] = xr.DataArray(
        data=face_nodes,
        coords={
            "face_x": ("face", face_x),
            "face_y": ("face", face_y),
        },
        dims=["face", "nmax_face"],
        attrs={
            "cf_role": "face_node_connectivity",
            "long_name": "Vertex nodes of mesh faces (counterclockwise)",
            "start_index": 0,
            "_FillValue": -1,
        },
    )
    ds.attrs = {"Conventions": "CF-1.8 UGRID-1.0"}
    return ds


def ugrid2d_topology(data: Union[xr.DataArray, xr.Dataset]) -> xr.Dataset:
    """
    Derive the 2D-UGRID quadrilateral mesh topology from a structured DataArray
    or Dataset, with (2D-dimensions) "y" and "x".

    Parameters
    ----------
    data: Union[xr.DataArray, xr.Dataset]
        Structured data from which the "x" and "y" coordinate will be used to
        define the UGRID-2D topology.

    Returns
    -------
    ugrid_topology: xr.Dataset
        Dataset with the required arrays describing 2D unstructured topology:
        node_x, node_y, face_x, face_y, face_nodes (connectivity).
    """
    # Transform midpoints into vertices
    # These are always returned monotonically increasing
    x = data["x"].values
    xcoord = _coord(data, "x")
    if not data.indexes["x"].is_monotonic_increasing:
        xcoord = xcoord[::-1]
    y = data["y"].values
    ycoord = _coord(data, "y")
    if not data.indexes["y"].is_monotonic_increasing:
        ycoord = ycoord[::-1]
    # Compute all vertices, these are the ugrid nodes
    node_y, node_x = (a.ravel() for a in np.meshgrid(ycoord, xcoord, indexing="ij"))
    face_y, face_x = (a.ravel() for a in np.meshgrid(y, x, indexing="ij"))
    linear_index = np.arange(node_x.size, dtype=np.int).reshape(
        ycoord.size, xcoord.size
    )
    # Allocate face_node_connectivity
    nfaces = (ycoord.size - 1) * (xcoord.size - 1)
    face_nodes = np.empty((nfaces, 4))
    # Set connectivity in counterclockwise manner
    face_nodes[:, 0] = linear_index[:-1, 1:].ravel()  # upper right
    face_nodes[:, 1] = linear_index[:-1, :-1].ravel()  # upper left
    face_nodes[:, 2] = linear_index[1:, :-1].ravel()  # lower left
    face_nodes[:, 3] = linear_index[1:, 1:].ravel()  # lower right
    # Tie it together
    ds = _ugrid2d_dataset(node_x, node_y, face_x, face_y, face_nodes)
    return ds


def ugrid2d_data(da: xr.DataArray) -> xr.DataArray:
    """
    Reshape a structured (x, y) DataArray into unstructured (face) form.
    Extra dimensions are maintained:
    e.g. (time, layer, x, y) becomes (time, layer, face).

    Parameters
    ----------
    da: xr.DataArray
        Structured DataArray with last two dimensions ("y", "x").

    Returns
    -------
    Unstructured DataArray with dimensions ("y", "x") replaced by ("face",).
    """
    if da.dims[:-2] == ("y", "x"):
        raise ValueError('Last two dimensions must be ("y", "x").')
    extra_dims = list(set(da.dims) - set(["y", "x"]))
    shape = da.data.shape
    new_shape = shape[:-2] + (np.product(shape[-2:]),)
    try:
        return xr.DataArray(
            data=da.data.reshape(new_shape),
            coords={k: da[k] for k in da.coords if k not in ("y", "x", "dy", "dx")},
            dims=extra_dims + ["face"],
        )
    except ValueError:
        return xr.DataArray(
            data=da.data.reshape(new_shape),
            coords={k: da[k] for k in da.coords if k not in ("y", "x", "dy", "dx")},
            dims=extra_dims[::-1] + ["face"],
        )


def _unstack_layers(ds: xr.Dataset) -> xr.Dataset:
    """
    Unstack the layer dimensions, as MDAL does not have support for
    UGRID-2D-layered datasets yet. Layers are stored as separate variables
    instead for now.
    """
    for variable in ds.data_vars:
        if "layer" in ds[variable].dims:
            stacked = ds[variable]
            ds = ds.drop_vars(variable)
            for layer in stacked["layer"].values:
                ds[f"{variable}_layer_{layer}"] = stacked.sel(layer=layer)
    if "layer" in ds.dims:
        ds = ds.drop("layer")
    return ds


def to_ugrid2d(data: Union[xr.DataArray, xr.Dataset]) -> xr.Dataset:
    """
    Convert a structured DataArray or Dataset into its UGRID-2D quadrilateral
    equivalent.

    See:
    https://ugrid-conventions.github.io/ugrid-conventions/#2d-flexible-mesh-mixed-triangles-quadrilaterals-etc-topology

    Parameters
    ----------
    data: Union[xr.DataArray, xr.Dataset]
        Dataset or DataArray with last two dimensions ("y", "x").
        In case of a Dataset, the 2D topology is defined once and variables are
        added one by one.
        In case of a DataArray, a name is required; a name can be set with:
        ``da.name = "..."``'

    Returns
    -------
    ugrid2d_dataset: xr.Dataset
        The equivalent data, in UGRID-2D quadrilateral form.
    """
    ds = ugrid2d_topology(data)
    if isinstance(data, xr.DataArray):
        if data.name is None:
            raise ValueError(
                'A name is required for the DataArray. It can be set with ``da.name = "..."`'
            )
        ds[data.name] = ugrid2d_data(data)
    elif isinstance(data, xr.Dataset):
        for variable in data.data_vars:
            ds[variable] = ugrid2d_data(data[variable])
    else:
        raise TypeError("data must be xarray.DataArray or xr.Dataset")
    # MDAL expects dates encoded as float, not integer, which is what xarray defaults to.
    # This leads to the variables not even showing up.
    for dim in ds.dims:
        if np.issubdtype(ds[dim].dtype, np.datetime64):
            ds[dim].encoding["dtype"] = float
    return _unstack_layers(ds)



import argparse
import json
import os
import platform
import sys
from contextlib import contextmanager, redirect_stderr, redirect_stdout
from os import devnull
from pathlib import Path
from typing import Any

import gistim


@contextmanager
def suppress_stdout_stderr():
    """A context manager that redirects stdout and stderr to devnull"""
    with open(devnull, "w") as fnull:
        with redirect_stderr(fnull) as err, redirect_stdout(fnull) as out:
            yield (err, out)


def write_json_stdout(data):
    sys.stdout.write(json.dumps(data))
    sys.stdout.write("\n")
    sys.stdout.flush()


def get_configdir() -> Path:
    if platform.system() == "Windows":
        configdir = Path(os.environ["APPDATA"]) / "qgis-tim"
    else:
        configdir = Path(os.environ["HOME"]) / ".qgis-tim"
    return configdir


def write_configjson(path: Path, data: dict[str, Any]) -> None:
    """
    Create, overwrite if needed.
    """
    content = {sys.executable: data}
    with open(path, "w") as f:
        f.write(json.dumps(content))
    return


def write_versions():
    # Store version numbers
    import timml
    import ttim

    versions = {
        "timml": timml.__version__,
        "ttim": ttim.__version__,
        "gistim": gistim.__version__,
    }
    configdir = get_configdir()
    path = configdir / "tim-versions.json"
    write_configjson(path, versions)
    return


def configure(args) -> None:
    """
    Write all the environmental variables so the QGIS interpreter
    can (re)set them properly.
    """
    configdir = get_configdir()

    # Store enviromental variables
    configdir.mkdir(exist_ok=True)
    env = {key: value for key, value in os.environ.items()}
    path = configdir / "environmental-variables.json"
    write_configjson(path, env)
    write_versions()
    return


def handle(line) -> None:
    data = json.loads(line)
    # print("JSON received:")
    # print(json.dumps(data, indent=4))

    operation = data.pop("operation")

    if operation == "compute":
        gistim.compute(
            inpath=data["inpath"],
            outpath=data["outpath"],
            cellsize=data["cellsize"],
            mode=data["mode"],
            active_elements=data["active_elements"],
        )
        response = "Computation of {inpath} to {outpath}".format(**data)
    elif operation == "convert":
        inpath = data["inpath"]
        outpath = data["outpath"]
        gistim.convert_to_script(inpath, outpath)
        response = "Conversion of {inpath} to {outpath}".format(**data)
    elif operation == "extract":
        inpath = data["inpath"]
        outpath = data["outpath"]
        wkt_geometry = data["wkt_geometry"].split(";")
        gistim.data_extraction.netcdf_to_table(
            inpath=inpath,
            outpath=outpath,
            wkt_geometry=wkt_geometry,
        )
        response = "Extraction of {inpath} to {outpath}".format(**data)
    elif operation == "process_ID":
        response = os.getpid()
    else:
        response = (
            "Invalid operation. Valid options are: "
            '"compute", "convert", "extract", "process_ID".'
        )

    return response


def serve(_) -> None:
    """
    Spin up a process listening for calls messages from the QGIS plugin.
    """
    try:
        write_json_stdout({"success": True, "message": "Initialized Tim server"})
        for line in sys.stdin:
            try:
                with suppress_stdout_stderr():
                    message = handle(line)
                response = {"success": True, "message": message}

            except Exception as error:
                response = {"success": False, "message": str(error)}

            write_json_stdout(response)

    except Exception as error:
        write_json_stdout({"success": False, "message": str(error)})


def extract(args) -> None:
    """
    Extract layer input from a netcdf dataset
    """
    inpath = args.inpath[0]
    outpath = args.outpath[0]
    wkt_geometry = args.wkt[0].split(";")
    gistim.data_extraction.netcdf_to_table(inpath, outpath, wkt_geometry)


def convert(args) -> None:
    """
    Convert a Geopackage into a Python script.
    """
    inpath = args.inpath[0]
    outpath = args.outpath[0]
    gistim.convert_to_script(inpath, outpath)


def compute(args) -> None:
    jsonpath = args.jsonpath[0]

    with open(jsonpath, "r") as f:
        data = json.loads(f.read())

    gistim.compute(
        inpath=data["inpath"],
        outpath=data["outpath"],
        cellsize=data["cellsize"],
        mode=data["mode"],
        active_elements=data["active_elements"],
    )
    return


if __name__ == "__main__":
    # Setup argparsers
    parser = argparse.ArgumentParser(prog="gistim")
    subparsers = parser.add_subparsers(help="sub-command help")
    parser_configure = subparsers.add_parser("configure", help="configure help")
    parser_serve = subparsers.add_parser("serve", help="serve help")
    parser_extract = subparsers.add_parser("extract", help="extract help")
    parser_convert = subparsers.add_parser("convert", help="convert help")
    parser_compute = subparsers.add_parser("compute", help="compute help")

    parser_configure.set_defaults(func=configure)

    parser_serve.set_defaults(func=serve)

    parser_extract.set_defaults(func=extract)
    parser_extract.add_argument("inpath", type=str, nargs=1, help="inpath")
    parser_extract.add_argument("outpath", type=str, nargs=1, help="outpath")
    parser_extract.add_argument("wkt", type=str, nargs=1, help="wkt")

    parser_convert.set_defaults(func=convert)
    parser_convert.add_argument("inpath", type=str, nargs=1, help="inpath")
    parser_convert.add_argument("outpath", type=str, nargs=1, help="outpath")

    parser_compute.set_defaults(func=compute)
    parser_compute.add_argument("jsonpath", type=str, nargs=1, help="jsonpath")

    # Parse and call the appropriate function
    args = parser.parse_args()
    args.func(args)



import pathlib
from typing import Dict, Union

import pkg_resources

from gistim import timml_elements, ttim_elements
from gistim.common import gridspec, model_specification
from gistim.compute import compute_steady, compute_transient
from gistim.data_extraction import as_aquifer_aquitard, layer_statistics
from gistim.ugrid import to_ugrid2d

# version
try:
    __version__ = pkg_resources.get_distribution(__name__).version
except pkg_resources.DistributionNotFound:
    # package is not installed
    pass


def convert_to_script(inpath: str, outpath: str) -> None:
    timml_spec, ttim_spec = model_specification(inpath, {})
    timml_script = timml_elements.convert_to_script(timml_spec)
    try:
        ttim_script = ttim_elements.convert_to_script(ttim_spec)
    except Exception:
        ttim_script = ""

    with open(outpath, "w") as f:
        f.write(timml_script)
        f.write("\n")
        f.write(ttim_script)


def compute(
    inpath: Union[pathlib.Path, str],
    outpath: Union[pathlib.Path, str],
    mode: str,
    cellsize: float,
    active_elements: Dict[str, bool],
) -> None:
    """
    Compute the results of TimML model.

    The model is fully specified by the GeoPacakge dataset in the path.

    The extent of the head grids is read from a vector layer in the
    GeoPackage file.

    Parameters
    ----------
    path: Union[pathlib.Path, str]
        Path to the GeoPackage file containing the full model input.
    cellsize: float
        Grid cell size of the computed output

    Returns
    -------
    None
        The result is written to a netCDF file. Its name is generated from
        the geopackage name, and the requested grid cell size.
    """
    if mode == "steady-state":
        compute_steady(inpath, outpath, cellsize, active_elements)
    elif mode == "transient":
        compute_transient(inpath, outpath, cellsize, active_elements)
    else:
        raise ValueError(f'Invalid mode: {mode}: should be "steady" or "transient".')
    return
